// Entry point of the mobile TV system

#include "mtv.hpp"

// global mtv instance
Mtv mtv;

// default constructor
Mtv::Mtv() : confMgr("/tmp/bear.txt") {
    // init subcomponents
    timeslicer = new Timeslicer();
    for (int i = 0; i < NUM_ENCAP_THREAD; i++) {
        encapsulator[i] = new Encapsulator();
    }
    transmitter = new Transmitter();
    // create mutex
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    // pthread_mutexattr_settype(&attr,PTHREAD_MUTEX_RECURSIVE);  // don't need this
    pthread_mutex_init(&reqMutex, &attr);
    pthread_mutex_init(&rdyMutex, &attr);
    pthread_mutexattr_destroy(&attr);
}

Mtv::~Mtv() {
    delete transmitter;
    //delete encapsulator;
    for (int i = 0; i < NUM_ENCAP_THREAD; i++) {
        delete encapsulator[i];
    }
    delete timeslicer;
    // pending bursts
    while (reqQueue.size() > 0) {
        Burst* b = popReq();
        delete b;
    } 
    while (rdyQueue.size() > 0) {
        Burst* b = popRdy();
        delete b;
    } 
    pthread_mutex_destroy(&reqMutex);
    pthread_mutex_destroy(&rdyMutex);
}



// called by the timeslicer thread
void Mtv::pushReq(Burst* burst) {
    pthread_mutex_lock(&reqMutex);
    reqQueue.push(burst);
    pthread_mutex_unlock(&reqMutex);
}

// blocking call, the caller will sleep for POLL_INTERVAL msec if no data is available
// called by the encapsulatore thread(s)
Burst* Mtv::popReq() {
    struct timespec nanosleeptime;
    Burst* ret = NULL;
    while (ret == NULL) {
        while (reqQueue.size() <= 0) { 
            unsigned long long int end = Common::getCurTimeMicros() + POLL_INTERVAL; 
            nanosleeptime.tv_sec = end / 1000000;
            nanosleeptime.tv_nsec = end % 1000000;
            clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &nanosleeptime, NULL);  // ignore the return value
        }
        // there might be some data, lock the mutex now
        pthread_mutex_lock(&reqMutex);
        if (reqQueue.size() > 0) {
            ret = reqQueue.top();
            reqQueue.pop();
        }  // if we didn't get it, it's fine, go back to sleep
        pthread_mutex_unlock(&reqMutex);
    }
    return ret;
}

// called by the encapsulator thread(s)
void Mtv::pushRdy(Burst* burst) {
    pthread_mutex_lock(&rdyMutex);
    rdyQueue.push(burst);
    pthread_mutex_unlock(&rdyMutex);
}

// blocking call, the caller will sleep for POLL_INTERVAL msec if no data is available
// called by the transmitter thread
Burst* Mtv::popRdy() {
    struct timespec nanosleeptime;
    Burst* ret = NULL;
    while (ret == NULL) {
        while (rdyQueue.size() <= 0) { 
            unsigned long long int end = getCurTimeMicros() + POLL_INTERVAL; 
            nanosleeptime.tv_sec = end / 1000000;
            nanosleeptime.tv_nsec = end % 1000000;
            clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &nanosleeptime, NULL);  // ignore the return value
        }
        // there might be some data, lock the mutex now
        pthread_mutex_lock(&rdyMutex);
        if (rdyQueue.size() > 0) {
            ret = rdyQueue.top();
            rdyQueue.pop();
        }  // if we didn't get it, it's fine, go back to sleep
        pthread_mutex_unlock(&rdyMutex);
    }
    return ret;
}

void Mtv::run() {
    // fine tune the stack size
    size_t stacksize;
    pthread_attr_t pattr;
    pthread_attr_init(&pattr);
    pthread_attr_getstacksize (&pattr, &stacksize);
    LOG_DEBUG("default stack size: %d", stacksize);
    stacksize = THREAD_STACK_SIZE;
    LOG_DEBUG("new stack size: %d", stacksize);
    pthread_attr_setstacksize (&pattr, stacksize);
    // the timeslicer thread
    if (0 != pthread_create(&t_thread, &pattr, start_t_thread, NULL)) {
        LOG_WARN1("Unable to create timeslicer thread");
    }
    // the encapsulator threads
    for (int i = 0; i < NUM_ENCAP_THREAD; i++) {
        if (0 != pthread_create(&e_thread[i], &pattr, start_e_thread, (void*) i++)) {
            LOG_WARN("Unable to create encapsulator thread: #%d", i);
        }
    }
    if (0 != pthread_create(&r_thread, &pattr, start_r_thread, NULL)) {
        LOG_WARN1("Unable to create transmitter thread");
    }
    // I don't want to return
    pthread_join(t_thread, NULL);
    for (int i = 0; i < NUM_ENCAP_THREAD; i++) {
        pthread_join(e_thread[i], NULL); 
    }
    pthread_join(r_thread, NULL);
    LOG_DEBUG1("Run function exits"); // system exits
}

// here come a few wrapper functions
void* start_t_thread(void* pParam) {
    mtv.timeslicer->schedule();
    LOG_DEBUG1("Timeslicer thread exits");
    return 0;
}
void* start_e_thread(void* pParam) {
    int thread_index = (int)pParam;    
    mtv.encapsulator[thread_index]->encapsulate();
    LOG_DEBUG("Encapsulator thread #%d exits", thread_index);
    return 0;
}
void* start_r_thread(void* pParam) {
    mtv.transmitter->transmit();
    LOG_DEBUG1("Transmitter thread exits");
    return 0;
}


int main(int argc, char *argv[]) {
    confmgr.console(); // get into console interface
    // spawn threads ? have a console command?
    mtv.run();
    return 0;
}

