
#include "networkconf.hpp"

/**
 * For QPSK, 16-QAM, and 64-QAM, the estimated bitrates in Mbit/s for different 
 * code rates and guard intervals are given. Each line represents one code rate out of 
 * [1/2, 2/3, 3/4, 5/6, 7/8] and each column represents one guard interval out of
 * [1/4, 1/8, 1/16, 1/32]. It is assumed that the full bandwidth is available to DVB-H,
 * hence no other (non-timesliced) streams or PSI/SI streams are taken into account.
 *
 * Example: The estimated bitrate for OFDM with 16-QAM modulation, 8MHz bandwidth, 
 * code rate 2/3 and guard interval 1/32 is 16.09 Mbit/s.
 *
 * See ETSI EN 300 744 (v1.5.1), p. 59
 */
const float NetworkConf::BITRATES_5MHZ[] = {
        /* QPSK */
        3.110, 3.456, 3.659, 3.770,
        4.147, 4.608, 4.879, 5.027,
        4.665, 5.184, 5.489, 5.655,
        5.184, 5.760, 6.099, 6.283,
        5.443, 6.048, 6.404, 6.598,
        /* 16-QAM */
        6.221, 6.912, 7.318, 7.540,
        8.294, 9.216, 9.758, 10.053,
        9.331, 10.368, 10.978, 11.310,
        10.368, 11.520, 12.197, 12.567,
        10.886, 12.096, 12.807, 13.195,
        /* 64-QAM */
        9.331, 10.368, 10.978, 11.310,
        12.441, 13.824, 14.637, 15.080,
        13.996, 15.551, 16.466, 16.965,
        15.551, 17.279, 18.296, 18.850,
        16.329, 18.143, 19.211, 19.793};
const float NetworkConf::BITRATES_7MHZ[] = {
        /* QPSK */
        4.354, 4.838, 5.123, 5.278,
        5.806, 6.451, 6.830, 7.037,
        6.532, 7.257, 7.684, 7.917,
        7.257, 8.064, 8.538, 8.797,
        7.620, 8.467, 8.965, 9.237,
        /* 16-QAM */
        8.709, 9.676, 10.246, 10.556,
        11.612, 12.902, 13.661, 14.075,
        13.063, 14.515, 15.369, 15.834,
        14.515, 16.127, 17.076, 17.594,
        15.240, 16.934, 17.930, 18.473,
        /* 64-QAM */
        13.063, 14.515, 15.369, 15.834,
        17.418, 19.353, 20.491, 21.112,
        19.595, 21.772, 23.053, 23.751,
        21.772, 24.191, 25.614, 26.390,
        22.861, 25.401, 26.895, 27.710};
const float NetworkConf::BITRATES_8MHZ[] = {
        /* QPSK */
        4.98,  5.53,  5.85,  6.03,
        6.64,  7.37,  7.81,  8.04,
        7.46,  8.29,  8.78,  9.05,
        8.29,  9.22,  9.76,  10.05,
        8.71,  9.68,  10.25, 10.56,
        /* 16-QAM */
        9.95,  11.06, 11.71, 12.06,
        13.27, 14.75, 15.61, 16.09,
        14.93, 16.59, 17.56, 18.10,
        16.59, 18.43, 19.52, 20.11,
        17.42, 19.35, 20.49, 21.11,
        /* 64-QAM */
        14.93, 16.59, 17.56, 18.10,
        19.91, 22.12, 23.42, 24.13,
        22.39, 24.88, 26.35, 27.14,
        24.88, 27.65, 29.27, 30.16,
        26.13, 29.03, 30.74, 31.67};


// this function returns the number of bytes required by saving the network data
int NetworkConf::getDataLength() {
    int length = 0;
    length += sizeof(numProgram);
    length += (1 + networkName.length()); // one byte for null termination
    length += sizeof(networkId);
    length += sizeof(uint32_t); // framesize
    length += sizeof (uint32_t); // FEC
    length += sizeof(maxBurstDuration);
    length += sizeof(uint32_t); // avgrate
    length += sizeof(platformId);
    length += (1 + platformName.length());
    length += sizeof(uint32_t); // langCode;
    length += sizeof(transportStreamId);
    length += sizeof(intServiceId);
    length += sizeof(frequency);
    length += sizeof(uint32_t); // bandwidth;
    length += sizeof(uint32_t); // constellation;
    length += sizeof(uint32_t); // hierarchy;
    length += sizeof(uint32_t); // codeRateHp;
    length += sizeof(uint32_t); // codeRateLp;
    length += sizeof(uint32_t); // guardInterval;
    length += sizeof(uint32_t); // transmitMode;
    length += sizeof(reservedPsiBitRate);
    length += (1 + outTsFilename.length());
    length += sizeof(uint32_t); // scheduler;

    return length;
}

int NetworkConf::fillData(unsigned char *buf, int offset) {
    offset = uint8ToByteArray(buf, offset, numProgram);
    offset = stringToByteArray(buf, offset, networkName);
    offset = uint8ToByteArray(buf, offset, networkId);
    offset = enumToByteArray(buf, offset, framesize);
    offset = enumToByteArray(buf, offset, fec);
    offset = uint16ToByteArray(buf, offset, maxBurstDuration);
    offset = enumToByteArray(buf, offset, avgrate);
    offset = uint32ToByteArray(buf, offset, platformId);
    offset = stringToByteArray(buf, offset, platformName);
    offset = enumToByteArray(buf, offset, langCode);
    offset = uint32ToByteArray(buf, offset, transportStreamId);
    offset = uint16ToByteArray(buf, offset, intServiceId);  
    offset = uint32ToByteArray(buf, offset, frequency);
    offset = enumToByteArray(buf, offset, bandwidth);
    offset = enumToByteArray(buf, offset, constellation);
    offset = enumToByteArray(buf, offset, hierarchy);
    offset = enumToByteArray(buf, offset, codeRateHp);
    offset = enumToByteArray(buf, offset, codeRateLp);
    offset = enumToByteArray(buf, offset, guardInterval);
    offset = enumToByteArray(buf, offset, transmitMode);
    offset = uint32ToByteArray(buf, offset, reservedPsiBitRate);
    offset = stringToByteArray(buf, offset, outTsFilename);
    offset = enumToByteArray(buf, offset, scheduler);
    return offset;
}


int NetworkConf::readData(unsigned char *buf, int offset) {
    uint32_t enumeration;
    offset = byteArrayToUint8(buf, offset, numProgram);
    offset = byteArrayToString(buf, offset, networkName);
    offset = byteArrayToUint8(buf, offset, networkId);
    offset = byteArrayToEnum(buf, offset, enumeration); framesize = (FRAMESIZE) enumeration;
    offset = byteArrayToEnum(buf, offset, enumeration); fec = (FEC) enumeration;
    offset = byteArrayToUint16(buf, offset, maxBurstDuration);
    offset = byteArrayToEnum(buf, offset, enumeration); avgrate = (AVGRATE) enumeration;
    offset = byteArrayToUint32(buf, offset, platformId);
    offset = byteArrayToString(buf, offset, platformName);
    offset = byteArrayToEnum(buf, offset, enumeration); langCode = (LANGCODE) enumeration;
    offset = byteArrayToUint32(buf, offset, transportStreamId);
    offset = byteArrayToUint16(buf, offset, intServiceId);  
    offset = byteArrayToUint32(buf, offset, frequency);
    offset = byteArrayToEnum(buf, offset, enumeration); bandwidth = (BANDWIDTH) enumeration;
    offset = byteArrayToEnum(buf, offset, enumeration); constellation = (CONSTELLATION) enumeration;
    offset = byteArrayToEnum(buf, offset, enumeration); hierarchy = (HIERARCHY) enumeration;
    offset = byteArrayToEnum(buf, offset, enumeration); codeRateHp = (CODERATEHP) enumeration;
    offset = byteArrayToEnum(buf, offset, enumeration); codeRateLp = (CODERATELP) enumeration;
    offset = byteArrayToEnum(buf, offset, enumeration); guardInterval = (GUARDINTERVAL) enumeration;
    offset = byteArrayToEnum(buf, offset, enumeration); transmitMode = (TRANSMITMODE) enumeration;
    offset = byteArrayToUint32(buf, offset, reservedPsiBitRate);
    offset = byteArrayToString(buf, offset, outTsFilename);
    offset = byteArrayToEnum(buf, offset, enumeration); scheduler = (SCHEDULER) enumeration;
 
    return offset;
}

string NetworkConf::toString() {
    string ret = "";
    ret += ("\t" + networkfieldmap_[FNUMPROGRAM] + ": "); 
    ret += (stringify(numProgram) + "\n");
    ret += ("\t" + networkfieldmap_[FNETWORKNAME] + ": "); 
    ret += (networkName + "\n");
    ret += ("\t" + networkfieldmap_[FNETWORKID] + ": "); 
    ret += (stringify(networkId) + "\n");
    ret += ("\t" + networkfieldmap_[FFRAMESIZE] + ": "); 
    ret += (framesizemap_[framesize] + "\n");
    ret += ("\t" + networkfieldmap_[FFEC] + ": "); 
    ret += (fecmap_[fec] + "\n");
    ret += ("\t" + networkfieldmap_[FMAXBURSTDURATION] + ": "); 
    ret += (stringify(maxBurstDuration) + "\n");
    ret += ("\t" + networkfieldmap_[FAVGRATE] + ": "); 
    ret += (avgratemap_[avgrate] + "\n");
    ret += ("\t" + networkfieldmap_[FPLATFORMID] + ": "); 
    ret += (stringify(platformId) + "\n");
    ret += ("\t" + networkfieldmap_[FPLATFORMNAME] + ": "); 
    ret += (platformName + "\n");
    ret += ("\t" + networkfieldmap_[FLANGCODE] + ": "); 
    ret += (langcodemap_[langCode] + "\n");
    ret += ("\t" + networkfieldmap_[FTRANSPORTSTREAMID] + ": "); 
    ret += (stringify(transportStreamId) + "\n");
    ret += ("\t" + networkfieldmap_[FINTSERVICEID] + ": "); 
    ret += (stringify(intServiceId) + "\n");
    ret += ("\t" + networkfieldmap_[FFREQUENCY] + ": "); 
    ret += (stringify(frequency) + "\n");
    ret += ("\t" + networkfieldmap_[FBANDWIDTH] + ": "); 
    ret += (bandwidthmap_[bandwidth] + "\n");
    ret += ("\t" + networkfieldmap_[FCONSTELLATION] + ": "); 
    ret += (constellationmap_[constellation] + "\n");
    ret += ("\t" + networkfieldmap_[FHIERARCHY] + ": "); 
    ret += (hierarchymap_[hierarchy] + "\n");
    ret += ("\t" + networkfieldmap_[FCODERATEHP] + ": "); 
    ret += (coderatehpmap_[codeRateHp] + "\n");
    ret += ("\t" + networkfieldmap_[FCODERATELP] + ": "); 
    ret += (coderatehpmap_[codeRateLp] + "\n");
    ret += ("\t" + networkfieldmap_[FGUARDINTERVAL] + ": "); 
    ret += (guardintervalmap_[guardInterval] + "\n");
    ret += ("\t" + networkfieldmap_[FTRANSMITMODE] + ": "); 
    ret += (transmitmodemap_[transmitMode] + "\n");
    ret += ("\t" + networkfieldmap_[FRESERVEDPSIBITRATE] + ": "); 
    ret += (stringify(reservedPsiBitRate) + "\n");
    ret += ("\t" + networkfieldmap_[FOUTTSFILENAME] + ": "); 
    ret += (outTsFilename + "\n");
    ret += ("\t" + networkfieldmap_[FSCHEDULER] + ": ");
    ret += (schedulermap_[scheduler] + "\n");
    return ret;
}

STATUS NetworkConf::set() {
    // iterate through the NETWORKFIELD enum
    for (int i = NETWORKFIELDFIRST + 1; i < NETWORKFIELDLAST; i++) { 
        cout << i << ") " << networkfieldmap_[i] << endl;
    }
    uint32_t n;  // choice
    cout << "Choice? ";
    cin >> n; 
    cin.ignore(); // to clear up the CR in buffer
    cout << "Input " << networkfieldmap_[n] << ": "; 
    switch (n) {
        case FNUMPROGRAM: cout << "Readonly field" << endl; break;
        case FNETWORKNAME: getline(cin, networkName); break;
        case FNETWORKID: cin >> networkId; break;
        case FFRAMESIZE: for (int i = FRAMESIZEFIRST + 1; i < FRAMESIZELAST; i++) cout << i << ") " << framesizemap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= FRAMESIZEFIRST + 1 && n < FRAMESIZELAST) framesize = (FRAMESIZE) n; break;
        case FFEC: for (int i = FECFIRST + 1; i < FECLAST; i++) cout << i << ") " << fecmap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= FECFIRST + 1 && n < FECLAST) fec = (FEC) n; break;
        case FMAXBURSTDURATION: cin >> maxBurstDuration; break;
        case FAVGRATE: for (int i = AVGRATEFIRST + 1; i < AVGRATELAST; i++) cout << i << ") " << avgratemap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= AVGRATEFIRST + 1 && n < AVGRATELAST) avgrate = (AVGRATE) n; break; 
        case FPLATFORMID: cin >> platformId; break;
        case FPLATFORMNAME: getline(cin, platformName); break;
        case FLANGCODE: for (int i = LANGCODEFIRST + 1; i < LANGCODELAST; i++) cout << i << ") " << langcodemap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= LANGCODEFIRST + 1 && n < LANGCODELAST) langCode = (LANGCODE) n; break; 
        case FTRANSPORTSTREAMID: cin >> transportStreamId; break;
        case FINTSERVICEID: cin >> intServiceId; break;
        case FFREQUENCY: cin >> frequency; break;
        case FBANDWIDTH: for (int i = BANDWIDTHFIRST + 1; i < BANDWIDTHLAST; i++) cout << i << ") " << bandwidthmap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= BANDWIDTHFIRST + 1 && n < BANDWIDTHLAST) bandwidth = (BANDWIDTH) n;  break; 
        case FCONSTELLATION: for (int i = CONSTELLATIONFIRST + 1; i < CONSTELLATIONLAST; i++) cout << i << ") " << constellationmap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= CONSTELLATIONFIRST + 1 && n < CONSTELLATIONLAST) constellation = (CONSTELLATION) n;  break; 
        case FHIERARCHY: for (int i = HIERARCHYFIRST + 1; i < HIERARCHYLAST; i++) cout << i << ") " << hierarchymap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= HIERARCHYFIRST + 1 && n < HIERARCHYLAST) hierarchy = (HIERARCHY) n; break; 
        case FCODERATEHP: for (int i = CODERATEHPFIRST + 1; i < CODERATEHPLAST; i++) cout << i << ") " << coderatehpmap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= CODERATEHPFIRST + 1 && n < CODERATEHPLAST) codeRateHp = (CODERATEHP) n; break; 
        case FCODERATELP: for (int i = CODERATELPFIRST + 1; i < CODERATELPLAST; i++) cout << i << ") " << coderatelpmap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= CODERATELPFIRST + 1 && n < CODERATELPLAST) codeRateLp = (CODERATELP) n; break; 
        case FGUARDINTERVAL: for (int i = GUARDINTERVALFIRST + 1; i < GUARDINTERVALLAST; i++) cout << i << ") " << guardintervalmap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= GUARDINTERVALFIRST + 1 && n < GUARDINTERVALLAST) guardInterval = (GUARDINTERVAL) n; break; 
        case FTRANSMITMODE: for (int i = TRANSMITMODEFIRST + 1; i < TRANSMITMODELAST; i++) cout << i << ") " << transmitmodemap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= TRANSMITMODEFIRST + 1 && n < TRANSMITMODELAST) transmitMode = (TRANSMITMODE) n; break; 
        case FRESERVEDPSIBITRATE: cin >> reservedPsiBitRate; break;
        case FOUTTSFILENAME: getline(cin, outTsFilename); break;
        case FSCHEDULER: for (int i = SCHEDULERFIRST + 1; i < SCHEDULERLAST; i++) cout << i << ") " << schedulermap_[i] << ", "; cout << "Choice? "; cin >> n; cin.ignore();  if (n >= SCHEDULERFIRST + 1 && n < SCHEDULERLAST) scheduler = (SCHEDULER) n; break; 
    }
    return OK;
}

// Computes the totally available bandwidth, depending on the physical layer
// settings (OFDM bandwidth, guard interval, OFDM constellation and coderate.
//
// This method returns the total bitrate in bit/s that is available on the
// channel. It returns -1 if lookup fails
int NetworkConf::getNetBitrate() {
    short column, row;
    short numBlock;
    const float *tableP;
    float bitrateMbitPerS;
    unsigned int bitrateBitPerS;

    LOG_DEBUG1("Calculating send time per byte ...");
    LOG_DEBUG("Transmission parameters: guard interval=%s, bandwidth=%s, coderate=%s, constellation=%s", 
              guardintervalmap_[guardInterval], bandwidthmap_[bandwidth],
              coderatehpmap_[codeRateLp], constellationmap_[constellation]);
    switch (guardInterval) {
        case GI4:  column = 0; break;
        case GI8:  column = 1; break;
        case GI16: column = 2; break;
        case GI32: column = 3; break;
        default: 
            LOG_WARN("invalid guard interval %s", guardintervalmap_[guardInterval]);
            return -1;  // we give up
    } 
    switch (codeRateHp) {
        case CR2: row = 0; break;
        case CR3: row = 1; break;
        case CR4: row = 2; break;
        case CR6: row = 3; break;
        case CR8: row = 4; break;
        default:
            LOG_WARN("invalid code rate: %s", coderatehpmap_[codeRateLp]);
            return -1;  // we give up
    }
    switch (bandwidth) {
        case MHZ5: tableP = BITRATES_5MHZ; break;
        case MHZ7: tableP = BITRATES_7MHZ; break;
        case MHZ8: tableP = BITRATES_8MHZ; break;
        default:
            LOG_WARN("invalid bandwidth_: %s", bandwidthmap_[bandwidth]);
            return -1;  // we give up
    }
    switch(constellation) {
        case QPSK: numBlock = 0; break;
        case QAM16: numBlock = 1; break;
        case RAM64: numBlock = 2; break;
        default:
            LOG_WARN("invalid modulation: %s", constellationmap_[constellation]);
            return -1;  // we give up
    }
    // compute the bit rate per second
    bitrateMbitPerS = tableP[(20 * numBlock) + (4 * row) + column]; // expected net bitrate in mbit/s
    bitrateBitPerS = bitrateMbitPerS * (float) pow(10.0, 6.0);
    LOG_DEBUG("Total bitrate is %u bit/s", bitrateBitPerS);
    return bitrateBitPerS;
}

   
// This function computes the bitrate that is available for DVB-H services. The
// result will always be smaller or equal than the total available bandwidth,
// depending on the amount of bitrate that was allocated for other services via
// the reservedPsiBitRate value
//
// per the variables guardInt, coderate, modulation, bandwidth,
// netBitrate and reservedPsiBitRate have to be set return the available bitrate
// in bit/s on the channel return -1 if something is wrong
int NetworkConf::getAvailableBitrate() {
    unsigned int bitrateBitPerS = getNetBitrate();
  
    if ((netBitrate > reservedPsiBitRate)) {
        bitrateBitPerS -= reservedPsiBitRate;
        LOG_DEBUG("Received bitrate for PSISI is %u bit/s", reservedPsiBitRate);
        LOG_DEBUG("available bitrate is %u bit/s", bitrateBitPerS);
    }
    else {
        LOG_WARN1("bitrate offset is too high, no bitrate remaining");
        return -1;  // we give up
    }
    return bitrateBitPerS;
}

// Computes the time (in nanosecond) needed to send a single byte, given that
// only the bandwidth that is reserved for DVB-H services is available.
//
// the variables guardInt, coderate, modulation, bandwidth, netBitrate and
// reservedPsiBitRate have to be set return the time needed to send a single byte in
// nanoseconds
//
int NetworkConf::getAvailableTimePerByte() {
    unsigned int netBitrate = getNetBitrate();
    if (netBitrate < 0) return -1;
    int availableBitrate = get_available_bitrate();
    if (availableBitrate < 0) return -1;
    unsigned int timePerByte = (8 * pow(10.0, 9.0) / availableBitrate); // in nanoseconds
    LOG_DEBUG("absolute time per byte is %d ns", timePerByte);
    return timePerByte;
}

// Computes the time (in nanosecond) needed to send a single byte, given that all bandwidth on the channel is available.
//
// the variables guardInt, coderate, modulation, bandwidth, netBitrate and reservedPsiBitRate have to be set
// return the time needed to send a single byte in nanoseconds
int NetworkConf::getNetTimePerByte() {
    unsigned int netBitrate = getNetBitrate();
    if (netBitrate < 0) return -1;
    unsigned int timePerByte = (8 * pow(10.0, 9.0) / netBitrate); // in nanoseconds
    LOG_DEBUG("net time per byte is %d ns", timePerByte);
    return timePerByte;
}

    

